#!/usr/bin/env python

#udp server for the LED-cube
#(c) 2006 moritz hassert


import sys
import os
import socket
import time
import threading
import serial
import re
import random


##########CONSTANTS##########

DEFAULT_UDPPORT = 5282
DEFAULT_COMPORT = 0
DEFAULT_ANI_DIR = './daten/'

BUF_SIZE = 65536

MAX_IDLE_TIME = 60.0

FRAME_TIME = 0.09

FRAME_SIZE = 125

AUTHOR_SIZE = 50
TITLE_SIZE = 50

ANIMATION_TIME = 120.0

STATE_IDLE   = 0
STATE_FRAMES = 1



#message format:
#
# line 1: PACKET_MAGIC MSG_TYPE
# lines x messsage data

PACKET_MAGIC = 'CUBE_MODELLER_MSG'

MSG_TYPE_REQUEST_CUBE     = 'GO_ONLINE'

MSG_TYPE_RELEASE_CUBE     = 'GO_OFFLINE'

MSG_TYPE_SHOW_FRAME       = 'FRAME'
#message data: frame data (125 chars)

MSG_TYPE_UPLOAD_ANIMATION = 'ANIMATION'
#message data:
#AUTHOR
#TITLE
#frame-lines


KNOWN_MSG_TYPES = [MSG_TYPE_REQUEST_CUBE, MSG_TYPE_RELEASE_CUBE,
                   MSG_TYPE_SHOW_FRAME, MSG_TYPE_UPLOAD_ANIMATION]




class Animation:
    position = 0
    
    def load_from_file(self, filename):
        try:
            s = open(filename).read()
        except:
            return False
        else:
            return self.load_from_string(s)
        
    def load_from_string(self, s):
        self.reset()
        self.frames = s.splitlines()
        return (len(self.frames)>0)
        
    def reset(self):
        self.position = 0

    def next_frame(self):
        if self.position >= len(self.frames):
            self.position = 0
        frame = self.frames[self.position]
        self.position += 1
        return frame
        



class CubeServer( threading.Thread):
    sock = None
    com = None
    current_owner = None
    state = STATE_IDLE
    terminated = False

    def __init__(self, udpport=DEFAULT_UDPPORT, comport=DEFAULT_COMPORT, animation_dir=DEFAULT_ANI_DIR):
        threading.Thread.__init__(self)

        self.udpport = udpport
        self.comport = comport
        self.animation_dir = animation_dir + '/'

    def __del__(self):
        print "server object destroyed"
        if self.sock:
            self.sock.close()
        if self.com:
            self.com.close()

    def log_client(self, client, text):
        print 'client '+str(client[0])+':'+str(client[1])+': '+text
    

    def store_animation(self, client, author, title, data):
        date = time.strftime('%Y-%m-%d_%H-%M-%S')
        author = author.replace("/", "_")
        title = title.replace("/","_")
        filename = date+'_'+author[:AUTHOR_SIZE]+'_'+title[:TITLE_SIZE]+'.raw'
        num = 0
        while os.path.exists(self.animation_dir + filename):
            num += 1
            filename = date+'+'+str(num)+'_'+author+'_'+title+'.raw'
        
        open(self.animation_dir + filename, 'w').write(data)
        self.log_client(client, 'sent animation. stored as "'+filename+'".')
                


    def load_random_animation(self):
        files = os.listdir(self.animation_dir)

        if len(files)==0:
            return None
        
        file = random.choice(files)
            
        ani = Animation()
        if ani.load_from_file(self.animation_dir + file):
            print 'loaded animation "'+file+'".'
            return ani
        else:
            return None


    def show_frame(self, frame):
        frame = frame.strip()

        if len(frame) > FRAME_SIZE:
            print 'show_frame: frame too long. truncated'
            
        self.com.write(frame[:FRAME_SIZE])
        self.com.write("\n")




    def receive(self, timeout):
        self.sock.settimeout(timeout)
        try:
            data, addr = self.sock.recvfrom(BUF_SIZE)
        except socket.timeout: #catch only timeout, pass othern on
            return False, None, None
        else:
            return True, data, addr

    def wait_for_msg(self):
        return self.receive(None)

    def wait_for_msg_or_idle(self):
        return self.receive(MAX_IDLE_TIME)
    
    def test_for_msg(self):
        return self.receive(0.001)

    def send_offline_notice(self, addr):
        self.sock.sendto( PACKET_MAGIC+' '+MSG_TYPE_RELEASE_CUBE, addr)



    def do_request_cube(self, client):
        #set this client as new owner, inform old owner if not the same
        self.log_client(client, ' went online.')
        if self.current_owner and (self.current_owner!=client):
            self.send_offline_notice(self.current_owner)
            self.log_client(self.current_owner, 'has been informed.')
        self.current_owner = client
        self.state = STATE_FRAMES
    
    def do_release_cube(self, client):
        #release only if this client is still the owner
        if self.current_owner == client:
            self.log_client(client, 'went offline.')
            self.current_owner = None
            self.state = STATE_IDLE
        else:
            self.log_client(client, 'went offline. not owner.')
    
    def do_show_frame(self, client, frame):
        if len(frame)<1:
            self.log_client(client, 'sent empty frame message. dropped.')
            return

        #pass frame on for displaying if client is owner
        if self.current_owner == client:
            self.log_client(client, 'sent frame. show it.')
            # self.log_client(client, frame)
            self.show_frame(frame)
            self.state = STATE_FRAMES
        else:
            self.log_client(client, 'sent frame. drop it. not owner.')

    def do_upload_animation(self, client, ani):
        #store animation in repository
        
        decoded = ani.split('\n', 2)
        if len(decoded)!=3:
            self.log_client(client, 'sent broken animation. dropped')
            return

        author, title, data = decoded
        self.store_animation(client, author, title, data)
    

    def decode_message(self, client, data):

        msg = data.split('\n', 1)
        header = msg[0].split(' ')

        if (len(header)<2)or(header[0]!=PACKET_MAGIC)or(not (header[1] in KNOWN_MSG_TYPES))or(len(msg)<2):
            self.log_client(client, 'sent unrecognized or broken message. dropped.')
            return

        body   = msg[1]
        message_type = header[1]

        if message_type == MSG_TYPE_REQUEST_CUBE:
            self.do_request_cube(client)

        elif message_type == MSG_TYPE_RELEASE_CUBE:
            self.do_release_cube(client)

        elif message_type == MSG_TYPE_SHOW_FRAME:
            self.do_show_frame(client, body)

        elif message_type == MSG_TYPE_UPLOAD_ANIMATION:
            self.do_upload_animation(client, body)


    def server_loop(self):
        self.terminated = False
        
        while not self.terminated:
            try:
                if self.state == STATE_FRAMES:
                    #handle client
                    print 'showing client frames ...'
                    while (self.state == STATE_FRAMES) and (not self.terminated):
                        newdata, data, addr = self.wait_for_msg_or_idle()

                        if newdata:
                            #print 'received "',data,'" from ',addr
                            self.decode_message(addr, data)
                        else:
                           #timed out
                           self.state = STATE_IDLE
                           #self.send_offline_notice(self.current_owner)
                           #self.current_owner = None

                else:
                    print 'idle. playing an animation now ...'

                    while (self.state == STATE_IDLE) and (not self.terminated):
                
                        #load an animation
                        animation = self.load_random_animation()

                        if not animation:
                            #if no animation available, wait infinite for a msg
                            print 'no animation found. waiting for clients ...'
                            newdata, data, addr = self.wait_for_msg()
                            if newdata:
                                self.decode_message(addr, data)
                    
                        else:
                            #show it
                            started_at = time.time()
                        
                            while (self.state == STATE_IDLE) and (not self.terminated) and ((time.time()-started_at)<=ANIMATION_TIME):
                                self.show_frame(animation.next_frame())
                                time.sleep(FRAME_TIME)

                                newdata, data, addr = self.test_for_msg()
                                if newdata:
                                    self.decode_message(addr, data)


                                    
                                
            except KeyboardInterrupt:
                print "Ctrl-C pressed. terminating server."
                self.terminated = True
                
#            except Exception, detail:
#                print 'Exception in mainloop: "',detail,'"'
#                self.terminated = True



    def run(self):
        try:

            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.bind(('', self.udpport))

            self.com = serial.Serial(self.comport,57600,8,serial.PARITY_NONE,serial.STOPBITS_TWO,None,0,0)
            self.com.open()

            self.server_loop()
        finally:
            print 'CubeServer is terminating ...'
            if self.sock:
                self.sock.close()
                self.sock = None
            if self.com:
                self.com.close()
                self.com = None
        
        
            
def main():
    if len(sys.argv) > 1:
        udpport = eval(sys.argv[1])
    else:
        udpport = DEFAULT_UDPPORT        

    if len(sys.argv) > 2:
        comport = sys.argv[2]
    else:
        comport = DEFAULT_COMPORT        

    server = CubeServer(udpport, comport)

    #start in main thread
    server.run()

    return 1



main()



def start():
    global svr
    svr = CubeServer()
    svr.start()


def stop():
    global svr
    svr.terminated = True
    client = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    client.sendto("foo", ('localhost', 5282))

def restart():
    stop()
    start()
    
